/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package unikl.disco.nc.gui;

import java.awt.Shape;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.util.logging.Logger;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.entity.EntityCollection;
import org.jfree.chart.labels.XYSeriesLabelGenerator;
import org.jfree.chart.plot.CrosshairState;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYItemRendererState;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.XYDataset;
import org.jfree.ui.RectangleEdge;
import org.jfree.util.ShapeUtilities;
import unikl.disco.math.Util;
import unikl.disco.nc.Curve;

class CurveLabelGenerator implements XYSeriesLabelGenerator {
	public String generateLabel(XYDataset dataset, int series) {
		if(dataset instanceof CurveDataset) {
			CurveDataset cds = (CurveDataset)dataset;
			if(cds.getSeries(series) instanceof Curve) {
				Curve c = (Curve)cds.getSeries(series);
				return c.getAlias();
			} else {
				return "not a curve";
			}
		} else {
			return "not a curve";
		}
	}
}
/**
 *
 * @author ngollan
 */
public class CurveRenderer extends XYLineAndShapeRenderer {
	private static Logger log = null;

	static double size = 6.0;
	static double delta = size / 2.0;

	private static Shape pointShape = new Ellipse2D.Double(-delta, -delta, size, size);
	private static Shape limitShape = new Arc2D.Double(0.0, -2.0*delta, size, 2.0*size, 110.0, 140.0, Arc2D.OPEN);
	private static Shape uptickShape = new Line2D.Double(0.0, delta, 0.0, 0.0);

	public CurveRenderer() {
		super(true, true);

		if(log == null) {
			log = Logger.getLogger("AffineCurveRenderer");
		}
		
		this.setLegendItemLabelGenerator(new CurveLabelGenerator());
		setDrawSeriesLineAsPath(false);
	}

	// override the necessary methods to handle point values

	@Override
	protected void drawPrimaryLine(XYItemRendererState state,
			java.awt.Graphics2D g2, XYPlot plot, XYDataset dataset, int pass,
			int series, int item, ValueAxis domainAxis, ValueAxis rangeAxis,
			java.awt.geom.Rectangle2D dataArea)
	{
		log.finer("Start line rendering...");

		if (item == 0) {
			log.finer("Item is 0");
			return;
		}

		if(!(dataset instanceof CurveDataset)) {
			Logger.getLogger("AffineCurveRenderer").warning("Defering rendering superclass");
			super.drawPrimaryLine(state, g2, plot, dataset, pass, series,
					item, domainAxis, rangeAxis, dataArea);
		}

		CurveDataset c = (CurveDataset)dataset;
		double r = c.getRate(series, item);
		if(!Util.isFinite(r)) {
			log.finer("Infinite rate");
			return;
		}

		// get the data point...
		double x0 = dataset.getXValue(series, item);
		double y0 = dataset.getYValue(series, item);
		if (Double.isNaN(y0) || Double.isNaN(x0)) {
			log.finer("NaN value! y0=" + y0 + ", x0=" + x0);
			return;
		}

		double x1, y1;
		if(dataset.getItemCount(series)-1 > item) {
			x1 = dataset.getXValue(series, item+1);
			y1 = y0 + r*(x1-x0);
		} else {
			x1 = domainAxis.getUpperBound();
			y1 = y0 + r*(x1-x0);
		}

		RectangleEdge xAxisLocation = plot.getDomainAxisEdge();
		RectangleEdge yAxisLocation = plot.getRangeAxisEdge();

		double transX0 = domainAxis.valueToJava2D(x0, dataArea, xAxisLocation);
		double transY0 = rangeAxis.valueToJava2D(y0, dataArea, yAxisLocation);

		double transX1 = domainAxis.valueToJava2D(x1, dataArea, xAxisLocation);
		double transY1 = rangeAxis.valueToJava2D(y1, dataArea, yAxisLocation);

		// only draw if we have good values
		if (Double.isNaN(transX0) || Double.isNaN(transY0)
			|| Double.isNaN(transX1) || Double.isNaN(transY1)) {
			log.finer("NaN trans value!");
			return;
		}

		PlotOrientation orientation = plot.getOrientation();
		if (orientation == PlotOrientation.HORIZONTAL) {
			state.workingLine.setLine(transY0, transX0, transY1, transX1);
		}
		else if (orientation == PlotOrientation.VERTICAL) {
			state.workingLine.setLine(transX0, transY0, transX1, transY1);
		}

		if (state.workingLine.intersects(dataArea)) {
			drawFirstPassShape(g2, pass, series, item, state.workingLine);
		}

		log.finer("Line rendering finished normally.");
	}

	@Override
	protected void drawPrimaryLineAsPath(XYItemRendererState state,
			java.awt.Graphics2D g2, XYPlot plot, XYDataset dataset, int pass,
			int series, int item, ValueAxis domainAxis, ValueAxis rangeAxis,
			java.awt.geom.Rectangle2D dataArea)
	{
		super.drawPrimaryLineAsPath(state, g2, plot, dataset, pass, series,
				item, domainAxis, rangeAxis, dataArea);
	}

	@Override
	protected void drawSecondaryPass(java.awt.Graphics2D g2, XYPlot plot,
			XYDataset dataset, int pass, int series, int item,
			ValueAxis domainAxis, java.awt.geom.Rectangle2D dataArea,
			ValueAxis rangeAxis, CrosshairState crosshairState,
			EntityCollection entities)
	{
		if(!(dataset instanceof CurveDataset)) {
			super.drawSecondaryPass(g2, plot, dataset, pass, series, item,
					domainAxis, dataArea, rangeAxis, crosshairState, entities);
		}

		CurveDataset cs = (CurveDataset)dataset;
		Shape entityArea = null;

		// get the data point...
		double x1 = dataset.getXValue(series, item);
		double y1 = dataset.getYValue(series, item);
		double yp1 = cs.getYPointValue(series, item);
		if((Double.isNaN(yp1) && Double.isNaN(y1)) || Double.isNaN(x1)) {
			return;
		}

		PlotOrientation orientation = plot.getOrientation();
		RectangleEdge xAxisLocation = plot.getDomainAxisEdge();
		RectangleEdge yAxisLocation = plot.getRangeAxisEdge();
		double transX1 = domainAxis.valueToJava2D(x1, dataArea, xAxisLocation);
		double transY1 = rangeAxis.valueToJava2D(y1, dataArea, yAxisLocation);
		double transYP1 = rangeAxis.valueToJava2D(yp1, dataArea, yAxisLocation);

		if(getItemShapeVisible(series, item)) {
			//TODO: implement proper point/limit shapes
			Shape y_shape = limitShape;
			Shape yp_shape = pointShape;

			if (orientation == PlotOrientation.HORIZONTAL) {
				y_shape = ShapeUtilities.createTranslatedShape(y_shape, transY1, transX1);
				yp_shape = ShapeUtilities.createTranslatedShape(yp_shape, transYP1, transX1);
			} else if (orientation == PlotOrientation.VERTICAL) {
				y_shape = ShapeUtilities.createTranslatedShape(y_shape, transX1, transY1);
				yp_shape = ShapeUtilities.createTranslatedShape(yp_shape, transX1, transYP1);
			}

			entityArea = y_shape;

			// paint the right limit if necessary
			if(y_shape.intersects(dataArea) && y1 != yp1) {
				g2.setPaint(getItemPaint(series, item));
				//g2.fill(y_shape);
				g2.draw(y_shape);
			}

			// paint the point position
			if(yp_shape.intersects(dataArea)) {
				g2.setPaint(getItemPaint(series, item));
				g2.setStroke(getItemOutlineStroke(series, item));
				g2.draw(yp_shape);
			}
		}

		double xx = transX1;
		double yy = transY1;
		if (orientation == PlotOrientation.HORIZONTAL) {
			xx = transY1;
			yy = transX1;
		}

		// draw the item label if there is one...
		if (isItemLabelVisible(series, item)) {
			drawItemLabel(g2, orientation, dataset, series, item, xx, yy,
					(y1 < 0.0));
		}

		int domainAxisIndex = plot.getDomainAxisIndex(domainAxis);
		int rangeAxisIndex = plot.getRangeAxisIndex(rangeAxis);
		updateCrosshairValues(crosshairState, x1, yp1, domainAxisIndex,
				rangeAxisIndex, transX1, transY1, plot.getOrientation());

		// add an entity for the item, but only if it falls within the data
		// area...
		if (entities != null && dataArea.contains(xx, yy)) {
			addEntity(entities, entityArea, dataset, series, item, xx, yy);
		}
	}
}
